use std::{collections::HashMap, sync::LazyLock};

use base64::{prelude::BASE64_STANDARD, Engine};
use tera::{to_value, Result as TeraResult, Value};
use time::OffsetDateTime;

pub static ICONS: LazyLock<HashMap<&str, String>> = LazyLock::new(|| {
    let mut map = HashMap::new();
    map.insert(
        "unknown",
        BASE64_STANDARD.encode(include_bytes!("../../icons/unknown.png")),
    );
    map.insert(
        "back",
        BASE64_STANDARD.encode(include_bytes!("../../icons/back.png")),
    );
    map.insert(
        "dir",
        BASE64_STANDARD.encode(include_bytes!("../../icons/dir.png")),
    );
    map.insert(
        "binary",
        BASE64_STANDARD.encode(include_bytes!("../../icons/binary.png")),
    );
    map.insert(
        "image",
        BASE64_STANDARD.encode(include_bytes!("../../icons/image2.png")),
    );
    map.insert(
        "compressed",
        BASE64_STANDARD.encode(include_bytes!("../../icons/compressed.png")),
    );

    map
});

pub fn iconize(value: &Value, _args: &HashMap<String, Value>) -> TeraResult<Value> {
    let icon = value.as_str().unwrap_or("unknown");

    let icon = match icon {
        "DIR" | "BACK" => "dir",
        "image/png" | "image/jpg" | "image/webp" => "image",
        "application/zip" => "compressed",
        "application/octet-stream" => "binary",
        _ => "unknown",
    };

    let icons = &*ICONS;
    let icon = if icons.contains_key(icon) {
        &icons[icon]
    } else {
        println!("Unknown filetype: {}", icon);
        &icons["unknown"]
    };

    let str = format!("<img src=\"data:image/png;base64,{}\" alt=\"\" />", icon);

    Ok(to_value(str).unwrap())
}

pub fn from_ico(value: &Value, _args: &HashMap<String, Value>) -> TeraResult<Value> {
    let Some(icon) = value.as_array() else {
        return iconize(
            &Value::String(String::from("")),
            &std::default::Default::default(),
        );
    };

    let icon = icon
        .iter()
        .map(|n| n.as_u64().unwrap() as u8)
        .collect::<Vec<_>>();

    let data = BASE64_STANDARD.encode(icon);

    let str = format!("<img src=\"data:image/png;base64,{}\" alt=\"\"/>", data);

    Ok(to_value(str).unwrap())
}

pub fn size(value: &Value, _args: &HashMap<String, Value>) -> TeraResult<Value> {
    let Some(value) = value.as_f64() else {
        return Ok(value.clone());
    };

    Ok(to_value(human_bytes::human_bytes(value)).unwrap())
}

pub fn time(value: &Value, _args: &HashMap<String, Value>) -> TeraResult<Value> {
    let time_format: std::vec::Vec<time::format_description::FormatItem<'_>> =
        time::format_description::parse("[day]-[month]-[year] [hour]:[minute]:[second]").unwrap();

    let Some(time) = value.as_array() else {
        return Ok(value.clone());
    };

    let time: Vec<i64> = time.iter().map(|v| v.as_i64().unwrap()).collect();

    let new_time: OffsetDateTime = OffsetDateTime::now_utc();
    let new_time = new_time.replace_year(time[0] as i32).unwrap();
    let new_time = new_time.replace_ordinal(time[1] as u16).unwrap();
    let new_time = new_time.replace_hour(time[2] as u8).unwrap();
    let new_time = new_time.replace_minute(time[3] as u8).unwrap();
    let new_time = new_time.replace_second(time[4] as u8).unwrap();
    let new_time = new_time.replace_nanosecond(time[5] as u32).unwrap();

    Ok(to_value(new_time.format(&time_format).unwrap()).unwrap())
}

pub fn md(value: &Value, _args: &HashMap<String, Value>) -> TeraResult<Value> {
    // parse value as markdown and return html
    if let Some(value) = value.as_str() {
        if let Ok(md) = markdown::to_html_with_options(value, &markdown::Options::gfm()) {
            return Ok(to_value(md).unwrap());
        }
    }

    Ok(to_value(value).unwrap())
}
